import os
import platform
import tempfile

import numpy as np
import pytest
from matplotlib.path import Path

from ..core.sequence import Sequence
from ..tools.phylo import Phylogeny, align_marker, fasttree, phylogeny


@pytest.fixture
def sequences():
    seq1 = Sequence(
        {
            "sequence": "ATCGATCGATCGATCGGGCGATCGATCG",
            "seq": "ATCGATCGATCGATCGGGCGATCGATCG",
            "sequence_id": "seq1",
            "name": "seq1",
        }
    )
    seq2 = Sequence(
        {
            "sequence": "ATCGATCGATCGATCGGGGGATCGATCG",
            "seq": "ATCGATCGATCGATCGGGGGATCGATCG",
            "sequence_id": "seq2",
            "name": "seq2",
        }
    )
    seq3 = Sequence(
        {
            "sequence": "ATCGCATGATCGTTCGGGCGATCGATCG",
            "seq": "ATCGCATGATCGTTCGGGCGATCGATCG",
            "sequence_id": "seq3",
            "name": "seq3",
        }
    )
    return [seq1, seq2, seq3]


@pytest.fixture
def fasta_string(sequences):
    return "\n".join([s.fasta for s in sequences])


@pytest.fixture
def fasta_file(fasta_string):
    fasta_file = tempfile.NamedTemporaryFile(delete=False)
    fasta_file.write(fasta_string.encode("utf-8"))
    fasta_file.close()
    return fasta_file


@pytest.fixture
def sequences_aa():
    seq1 = Sequence("MEEPQSDPSVEPPLSQETFSDLWKLLPENNVISPLPSQAMDDLMLSPDDIEQ", id="seq1")
    seq2 = Sequence("MEEPQSDPSVEPPLSQETFSDLWKLLPENNVLSPLPSQAMDDLMLSPDDIEQ", id="seq2")
    seq3 = Sequence("MEEPQSDPSVEPPLSQETFSDIWKLLPENNVISPLPSQAMDDLMLSPDDIEQ", id="seq3")
    return [seq1, seq2, seq3]


@pytest.fixture
def fasta_string_aa(sequences_aa):
    return "\n".join([s.fasta for s in sequences_aa])


@pytest.fixture
def fasta_file_aa(fasta_string_aa):
    fasta_file = tempfile.NamedTemporaryFile(delete=False)
    fasta_file.write(fasta_string_aa.encode("utf-8"))
    fasta_file.close()
    return fasta_file


@pytest.fixture
def root_sequence():
    return Sequence(sequence="ATCGATCGATCGATCGGGCGATCGATCG", id="root")


# ---------------------------
#       fasttree
# ---------------------------``


def test_fasttree_with_file_input(fasta_file):
    tree_file = tempfile.NamedTemporaryFile(delete=False)
    tree_path = fasttree(
        fasta_file.name, tree_file=tree_file.name, debug=True, quiet=False
    )
    assert os.path.isfile(tree_path)
    with open(tree_path, "r") as f:
        tree_string = f.read()
    assert tree_string.startswith("(")
    os.remove(tree_path)


def test_fasttree_with_string_input(fasta_string):
    tree_file = tempfile.NamedTemporaryFile(delete=False)
    tree_path = fasttree(
        fasta_string, tree_file=tree_file.name, debug=True, quiet=False
    )
    assert os.path.isfile(tree_path)
    with open(tree_path, "r") as f:
        tree_string = f.read()
    assert tree_string.startswith("(")
    os.remove(tree_path)


def test_fasttree_with_aa_file_input(fasta_file_aa):
    tree_file = tempfile.NamedTemporaryFile(delete=False)
    tree_path = fasttree(
        fasta_file_aa.name,
        tree_file=tree_file.name,
        is_aa=True,
        debug=True,
        quiet=False,
    )
    assert os.path.isfile(tree_path)
    with open(tree_path, "r") as f:
        tree_string = f.read()
    assert tree_string.startswith("(")
    os.remove(tree_path)


def test_fasttree_with_aa_string_input(fasta_string_aa):
    tree_file = tempfile.NamedTemporaryFile(delete=False)
    tree_path = fasttree(
        fasta_string_aa, tree_file=tree_file.name, is_aa=True, debug=True, quiet=False
    )
    assert os.path.isfile(tree_path)
    with open(tree_path, "r") as f:
        tree_string = f.read()
    assert tree_string.startswith("(")
    os.remove(tree_path)


def test_fasttree_with_custom_binary(fasta_file):
    test_bin_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    fasttree_bin = os.path.join(
        test_bin_dir, f"tests/bin/fasttree_{platform.system().lower()}"
    )
    tree_file = tempfile.NamedTemporaryFile(delete=False)
    tree_path = fasttree(
        fasta_file.name,
        tree_file=tree_file.name,
        fasttree_bin=fasttree_bin,
        debug=True,
        quiet=False,
    )
    assert os.path.isfile(tree_path)
    with open(tree_path, "r") as f:
        tree_string = f.read()
    assert tree_string.startswith("(")
    os.remove(tree_path)


# ---------------------------
#        phylogeny
# ---------------------------


def test_phylogeny_with_sequence_input(sequences):
    phylo = phylogeny(sequences)
    assert isinstance(phylo, Phylogeny)
    assert len(phylo.sequences) == 3


def test_phylogeny_with_file_input(fasta_file):
    phylo = phylogeny(fasta_file.name)
    assert isinstance(phylo, Phylogeny)
    assert len(phylo.sequences) == 3


def test_phylogeny_with_list_input(sequences):
    phylo = phylogeny(sequences)
    assert isinstance(phylo, Phylogeny)
    assert len(phylo.sequences) == 3


def test_phylogeny_with_name(fasta_file):
    name = "test_phylogeny_with_name"
    phylo = phylogeny(fasta_file.name, name=name)
    assert isinstance(phylo, Phylogeny)
    assert phylo.name == name


def test_phylogeny_with_root_sequence(fasta_file, root_sequence):
    phylo = phylogeny(fasta_file.name, root=root_sequence)
    assert isinstance(phylo, Phylogeny)
    assert phylo.root.id == root_sequence.id


def test_phylogeny_with_root_sequence_id(fasta_file):
    root_seq_id = "seq1"
    phylo = phylogeny(fasta_file.name, root=root_seq_id)
    assert isinstance(phylo, Phylogeny)
    assert phylo.root.id == root_seq_id


def test_phylogeny_with_cluster(fasta_file):
    phylo = phylogeny(fasta_file.name, cluster=True, clustering_algo="cdhit")
    assert isinstance(phylo, Phylogeny)
    assert len(phylo.sequences) == 3
    assert len(phylo.clusters) == 3


def test_phylogeny_with_clustering_threshold(fasta_file):
    clustering_threshold = 0.95
    phylo = phylogeny(
        fasta_file.name,
        cluster=True,
        clustering_threshold=clustering_threshold,
        clustering_algo="cdhit",
    )
    assert isinstance(phylo, Phylogeny)
    assert len(phylo.sequences) == 3
    assert len(phylo.clusters) == 2


def test_phylogeny_with_rename_dict(fasta_file):
    rename_dict = {"seq1": "new_seq1", "seq2": "new_seq2"}
    phylo = phylogeny(fasta_file.name, rename=rename_dict)
    assert isinstance(phylo, Phylogeny)
    phylo_ids = [s.id for s in phylo.sequences]
    assert "new_seq1" in phylo_ids
    assert "new_seq2" in phylo_ids
    assert "seq1" not in phylo_ids
    assert "seq2" not in phylo_ids


def test_phylogeny_with_rename_callable(fasta_file):
    def rename_func(old_name):
        if old_name == "seq1":
            return "new_seq1"
        else:
            return None

    phylo = phylogeny(fasta_file.name, rename=rename_func)
    assert isinstance(phylo, Phylogeny)
    phylo_ids = [s.id for s in phylo.sequences]
    assert "new_seq1" in phylo_ids
    assert "seq1" not in phylo_ids


def test_phylogeny_with_id_key(sequences):
    phylo = phylogeny(sequences, id_key="name")
    assert isinstance(phylo, Phylogeny)
    assert len(phylo.sequences) == 3


def test_phylogeny_with_sequence_key(sequences):
    phylo = phylogeny(sequences, sequence_key="seq")
    assert isinstance(phylo, Phylogeny)
    assert len(phylo.sequences) == 3


# ---------------------------
#      align_marker
# ---------------------------


def test_align_marker_with_center_alignment():
    marker = "o"
    halign = "center"
    valign = "middle"
    marker_path = align_marker(marker, halign=halign, valign=valign)
    assert isinstance(marker_path, Path)
    assert np.allclose(marker_path.vertices[:, 0].max(), 0.5, atol=0.05)
    assert np.allclose(marker_path.vertices[:, 0].min(), -0.5, atol=0.05)
    assert np.allclose(marker_path.vertices[:, 1].max(), 0.5, atol=0.05)
    assert np.allclose(marker_path.vertices[:, 1].min(), -0.5, atol=0.05)
    assert np.allclose(marker_path.vertices.mean(axis=0), [0, 0], atol=0.05)


def test_align_marker_with_left_alignment():
    marker = "o"
    halign = "left"
    valign = "middle"
    marker_path = align_marker(marker, halign=halign, valign=valign)
    assert isinstance(marker_path, Path)
    assert np.allclose(marker_path.vertices[:, 0].max(), 1, atol=0.05)
    assert np.allclose(marker_path.vertices[:, 0].min(), 0, atol=0.05)
    assert np.allclose(marker_path.vertices[:, 1].max(), 0.5, atol=0.05)
    assert np.allclose(marker_path.vertices[:, 1].min(), -0.5, atol=0.05)
    assert np.allclose(marker_path.vertices.mean(axis=0), [0.5, 0], atol=0.05)


def test_align_marker_with_right_alignment():
    marker = "o"
    halign = "right"
    valign = "middle"
    marker_path = align_marker(marker, halign=halign, valign=valign)
    assert isinstance(marker_path, Path)
    assert np.allclose(marker_path.vertices[:, 0].max(), 0, atol=0.05)
    assert np.allclose(marker_path.vertices[:, 0].min(), -1, atol=0.05)
    assert np.allclose(marker_path.vertices[:, 1].max(), 0.5, atol=0.05)
    assert np.allclose(marker_path.vertices[:, 1].min(), -0.5, atol=0.05)
    assert np.allclose(marker_path.vertices.mean(axis=0), [-0.5, 0], atol=0.05)


def test_align_marker_with_top_alignment():
    marker = "o"
    halign = "center"
    valign = "top"
    marker_path = align_marker(marker, halign=halign, valign=valign)
    assert isinstance(marker_path, Path)
    assert np.allclose(marker_path.vertices[:, 0].max(), 0.5, atol=0.05)
    assert np.allclose(marker_path.vertices[:, 0].min(), -0.5, atol=0.05)
    assert np.allclose(marker_path.vertices[:, 1].max(), 0, atol=0.05)
    assert np.allclose(marker_path.vertices[:, 1].min(), -1, atol=0.05)
    assert np.allclose(marker_path.vertices.mean(axis=0), [0, -0.5], atol=0.05)


def test_align_marker_with_bottom_alignment():
    marker = "o"
    halign = "center"
    valign = "bottom"
    marker_path = align_marker(marker, halign=halign, valign=valign)
    assert isinstance(marker_path, Path)
    assert np.allclose(marker_path.vertices[:, 0].max(), 0.5, atol=0.05)
    assert np.allclose(marker_path.vertices[:, 0].min(), -0.5, atol=0.05)
    assert np.allclose(marker_path.vertices[:, 1].max(), 1, atol=0.05)
    assert np.allclose(marker_path.vertices[:, 1].min(), 0, atol=0.05)
    assert np.allclose(marker_path.vertices.mean(axis=0), [0, 0.5], atol=0.05)



#!/usr/bin/env python
# filename: cloning.py
#
#
# Copyright (c) 2023 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from typing import Callable, Iterable, Optional, Union

from tqdm.auto import tqdm

from ..core.sequence import Sequence, codon_optimize

__all__ = ["build_synthesis_constructs"]


def build_synthesis_constructs(
    sequences: Union[Sequence, Iterable[Sequence], str],
    overhang_5: Optional[dict] = None,
    overhang_3: Optional[dict] = None,
    sequence_key: Optional[str] = None,
    id_key: Optional[str] = None,
    locus_key: Optional[str] = None,
    frame: Optional[int] = None,
    add_locus_to_name: bool = True,
    group_by_chain: bool = False,
    show_progress: bool = True,
    sort_func: Optional[Callable] = None,
):
    """
    Builds codon-optimized synthesis constructs, including Gibson overhangs suitable
    for cloning IGH, IGK and IGL variable region constructs into antibody expression
    vectors.


    .. seealso::
        | Thomas Tiller, Eric Meffre, Sergey Yurasov, Makoto Tsuiji, Michel C Nussenzweig, Hedda Wardemann
        | Efficient generation of monoclonal antibodies from single human B cells by single cell RT-PCR and expression vector cloning
        | *Journal of Immunological Methods* 2008, doi: 10.1016/j.jim.2007.09.017


    Parameters
    ----------
    sequences : ``Sequence``, str, or ``Iterable[Sequence]``
        A ``Sequence`` object, a sequence as a ``str``, or an ``Iterable`` of ``Sequence`` objects.

        .. note::
            If the provided sequences are nucleotide sequences, ``frame`` must also be provided
            so that codon optimization can be performed on the correct reading frame.

    overhang_5 : dict, optional
        A ``dict`` mapping the locus name to 5' Gibson overhangs. By default, Gibson
        overhangs corresponding to the expression vectors in Tiller et al, 2008:

            | **IGH:** ``catcctttttctagtagcaactgcaaccggtgtacac``
            | **IGK:** ``atcctttttctagtagcaactgcaaccggtgtacac``
            | **IGL:** ``atcctttttctagtagcaactgcaaccggtgtacac``

        To produce constructs without 5' Gibson overhangs, provide an empty dictionary.

    overhang_3 : dict, optional
        A ``dict`` mapping the locus name to 3' Gibson overhangs. By default, Gibson
        overhangs corresponding to the expression vectors in Tiller et al, 2008:

            | **IGH:** ``gcgtcgaccaagggcccatcggtcttcc``
            | **IGK:** ``cgtacggtggctgcaccatctgtcttcatc``
            | **IGL:** ``ggtcagcccaaggctgccccctcggtcactctgttcccgccctcgagtgaggagcttcaagccaacaaggcc``

        To produce constructs without 3' Gibson overhangs, provide an empty dictionary.

    sequence_key : str, default='sequence_aa'
        Field containing the sequence to be codon optimized. Default is ``'sequence_aa'`` if
        ``annotation_format == 'airr'`` or ``'vdj_aa'`` if ``annotation_format == 'json'``.
        Either nucleotide or amino acid sequences are acceptable.

    id_key : str, optional
        Field containing the name of the sequence. If not provided, ``sequence.id`` is used.

    locus_key : str, optional
        Field containing the name of the locus. If not provided, ``sequence["locus"]`` is used, which
        conforms with AIRR naming conventions.

        .. note::
            The data in ``locus_key`` should return one of the following:
                * IGH, IGK, or IGL (which are the names of the loci in AIRR naming conventions)
                * heavy, kappa, or lambda

    frame : int, default=1
        Reading frame to translate. Default is ``1``.

        .. note::
            This parameter is ignored if the input sequences are amino acid sequences.

    add_locus_to_name : bool, default=True
        If ``True``, the name of the sequence will be appended with the name of the locus, separated by an underscore.
        If ``False``, the name of the sequence will not be modified.

    group_by_chain : bool, default=False
        If ``True``, the output ``list`` of ``Sequence`` objects will be separated by chain (heavy first, light second).
        Within each chain group, sequence ordering will retained (either input order, or sorted by ``sort_func``).
        If ``False``, the output ``list`` of ``Sequence`` objects will not be grouped.

    in_place : bool, default=False
        If ``True``, the input ``Sequence`` objects will be returned with the optimized sequence
        populating the ``sequence.codon_optimized`` property. If ``False``, the optimized sequences
        will be returned as new ``Sequence`` objects.

    sort_func : Callable, optional
        A function to sort the output ``list`` of ``Sequence`` objects. By default,
        the output ``list`` is is in the same order as the input.


    Returns
    -------
    sequences : ``list`` of ``Sequence`` objects
        A ``list`` of ``Sequence`` objects. Each ``Sequence`` object has the following properties:

            | *id*: The sequence ID (either ``sequence.id`` or ``id_key``), to which the locus (``locus_key``) is appended if ``add_locus_to_name`` is ``True``.
            | *sequence*: The codon-optimized sequence, including Gibson overhangs.

        If ``sort_func`` is provided, the output ``list``  will be sorted by ``sort_func``, otherwise the output is in the same order as the input.

    """
    # get overhangs
    overhang_3 = overhang_3 if overhang_3 is not None else GIBSON3
    overhang_5 = overhang_5 if overhang_5 is not None else GIBSON5

    # process input
    if isinstance(sequences, (Sequence, str)):
        sequences = [sequences]
    sequences = [Sequence(s) for s in sequences]

    # sort
    if sort_func is not None:
        sequences = sorted(sequences, key=sort_func)
    if group_by_chain:
        locus_key = locus_key if locus_key is not None else "locus"
        heavies = [s for s in sequences if s[locus_key] in ["heavy", "IGH"]]
        lights = [s for s in sequences if s[locus_key] not in ["heavy", "IGH"]]
        sequences = heavies + lights

    # make codon-optimized synthesis constructs
    if show_progress:
        sequences = tqdm(sequences)
    optimized = []
    for sequence in sequences:
        seq = sequence[sequence_key] if sequence_key is not None else sequence.sequence
        name = sequence[id_key] if id_key is not None else sequence.id
        locus = sequence[locus_key] if locus_key is not None else sequence["locus"]
        if add_locus_to_name and locus is not None:
            name += f"_{locus}"
        opt_seq = codon_optimize(seq, frame=frame, as_string=True)
        synthesis_construct = (
            overhang_5.get(locus, "") + opt_seq + overhang_3.get(locus, "")
        )
        optimized.append(Sequence(synthesis_construct, id=name))
    return optimized


GIBSON5 = {
    "IGH": "catcctttttctagtagcaactgcaaccggtgtacac",
    "IGK": "atcctttttctagtagcaactgcaaccggtgtacac",
    "IGL": "atcctttttctagtagcaactgcaaccggtgtacac",
    "heavy": "catcctttttctagtagcaactgcaaccggtgtacac",
    "kappa": "atcctttttctagtagcaactgcaaccggtgtacac",
    "lambda": "atcctttttctagtagcaactgcaaccggtgtacac",
}

GIBSON3 = {
    "IGH": "gcgtcgaccaagggcccatcggtcttcc",
    "IGK": "cgtacggtggctgcaccatctgtcttcatc",
    "IGL": "ggtcagcccaaggctgccccctcggtcactctgttcccgccctcgagtgaggagcttcaagccaacaaggcc",
    "heavy": "gcgtcgaccaagggcccatcggtcttcc",
    "kappa": "cgtacggtggctgcaccatctgtcttcatc",
    "lambda": "ggtcagcccaaggctgccccctcggtcactctgttcccgccctcgagtgaggagcttcaagccaacaaggcc",
}


LOCUS_MAP = {
    "IGH": "heavy",
    "IGK": "kappa",
    "IGL": "lambda",
    "TRA": "alpha",
    "TRB": "beta",
    "TRD": "delta",
    "TRG": "gamma",
}



#!/usr/bin/env python
# filename: preprocess.py


#
# Copyright (c) 2025 Bryan Briney & Benjamin Nemoz
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


# import os
# import time
# import datetime
# from typing import Optional
# from tqdm.auto import tqdm

# import polars as pl
# from .. import Sequence
# from ..io import list_files, make_dir, to_fasta, from_polars, to_polars
# from ..tools import cluster, alignment


# __all__ = ['deduplicate', 'reduction']


# #########################################
# #                                       #
# #         Helper functions              #
# #                                       #
# #########################################


# def _deduplicate_sequences(
#     df: pl.DataFrame,
#     group_by_cols: list,
#     keep_read_numbers: bool
#     ):

#     if not keep_read_numbers:
#         return df.unique(subset=group_by_cols).sort("sequence")
#     return (
#         df.group_by(group_by_cols).agg(
#             pl.count().alias("count"),
#             pl.col("sequence_id").first(),
#         ).sort("sequence")
#     )


# def _write_fasta(
#     df: pl.DataFrame,
#     file_path: str,
#     keep_read_numbers: bool,
#     read_number_separator: str
#     ):

#     if keep_read_numbers:
#         df = df.with_columns(
#             (pl.col("sequence_id") + read_number_separator + pl.col("count").cast(pl.Utf8)).alias("sequence_id")
#         )
#     tuples_list = list(zip(df["sequence_id"].to_list(), df["sequence"].to_list()))
#     to_fasta(tuples_list, file_path)
#     print(f"Output written to FASTA file: {file_path}\n")


# def _process_chains(
#     step: str,
#     chains: pl.DataFrame,
#     min_cluster_size: int,
#     clustering_threshold: float,
#     consentroid: str,
#     cluster_sizes_separator: str,
#     keep_cluster_sizes: bool,
#     output_format: str,
#     debug: bool
# ):
#     """Internal function to generate a consentroid from a bin of sequences"""

#     consentroids = []
#     for vj in tqdm(chains['vj_bin'].unique().to_list(), desc=f'{step} chains processing'):
#         group = chains.filter(pl.col('vj_bin') == vj)
#         if group.height < min_cluster_size:
#             continue

#         group_sequences = from_polars(group)
#         clusters = cluster.cluster(group_sequences, threshold=clustering_threshold)
#         for c in clusters:
#             if c.size < min_cluster_size:
#                 continue
#             if consentroid == 'centroid':
#                 seq_id = f"{c.centroid.id}{cluster_sizes_separator}{c.size}" if keep_cluster_sizes else c.centroid.id
#                 consentroids.append(Sequence(c.centroid.sequence, id=seq_id))
#             elif consentroid == 'consensus':
#                 _consensus = alignment.make_consensus([s for s in c.sequences], algo='mafft', alignment_kwargs={'mafft_bin': 'mafft', 'threads': 1}, debug=debug)
#                 seq_id = f"{_consensus.id}{cluster_sizes_separator}{c.size}" if keep_cluster_sizes else _consensus.id
#                 consentroids.append(Sequence(_consensus.sequence, id=seq_id))

#         if debug:
#             print(f"\tProcessed {len(clusters)} clusters for {vj}, corresponding to {len(group_sequences)} sequences")

#     if debug:
#         print(f"\tTotal concentroids generated: {len(consentroids)}\n")

#     if output_format == 'fasta':
#         return consentroids

#     elif output_format == 'airr':
#         df = to_polars(consentroids)
#         if keep_cluster_sizes and len(df) != 0:
#             df = df.with_columns(
#                 pl.col("sequence_id")
#                 .str.split_exact(cluster_sizes_separator, 1)  # Split into a struct with two fields
#                 .alias("split_id")
#             ).with_columns(
#                 pl.col("split_id").struct.field("field_0").alias("sequence_id"),  # Extract the first field
#                 pl.col("split_id").struct.field("field_1").cast(pl.Utf8).alias("count")  # Extract the second field
#             ).drop("split_id")  # Drop the intermediate "split_id" column
#         return df


# def _process_chain_group(
#     chains: pl.DataFrame,
#     chain_type: str,
#     umi: bool,
#     min_cluster_size: int,
#     clustering_threshold: float,
#     consentroid: str,
#     cluster_sizes_separator: str,
#     keep_cluster_sizes: bool,
#     output_format: str,
#     debug: bool
# ):
#     # Generate vj_bin column
#     if umi:
#         chains = chains.with_columns((pl.col("v_gene") + "_" + pl.col("j_gene") + "_" + pl.col("umi")).alias("vj_bin"))
#     else:
#         chains = chains.with_columns((pl.col("v_gene") + "_" + pl.col("j_gene")).alias("vj_bin"))

#     return _process_chains(chain_type, chains, min_cluster_size, clustering_threshold, consentroid, cluster_sizes_separator, keep_cluster_sizes, output_format, debug)


# #########################################
# #                                       #
# #           Main functions              #
# #                                       #
# #########################################


# def deduplicate(project_folder: str,
#     output: Optional[str] = None,
#     output_format: str = 'fasta',
#     pool: bool = True,
#     umi: bool = False,
#     keep_read_numbers: bool = False,
#     read_number_separator: str = '|',
#     large_files: bool = False,
#     debug: bool = False
# ) -> None:
#     '''
#     A polyvalent tool for deduplication of assigned reads. This function takes as input the AIRR-compliant
#     tables and is specifically designed to handle extremely large files with a minimal footprint.

#     Parameters:
#     project_folder (str): Path to the project folder containing AIRR-compliant tables.
#     output (str, optional): Subdirectory for output files. Created if non-existent. Defaults to None.
#     output_format (str): Either "fasta" or "airr". Default is "fasta".
#     pool (bool, optional): If True, pool all samples together. Defaults to True.
#     umi (bool, optional): If True, use UMI for deduplication. Defaults to False.
#     keep_read_numbers (bool, optional): If True, read numbers will be added to sequence names. Defaults to False.
#     large_files (bool, optional): If True, optimize for large files (>100Go). Defaults to False.
#     debug (bool, optional): If True, print debug information. Defaults to False.

#     Returns: None. A FASTA- or TSV- file is written to disk.
#     '''

#     start = time.time()

#     # Assert that output_format is valid
#     if output_format not in ['fasta', 'airr']:
#         raise ValueError("Invalid output_format. Must be 'fasta' or 'airr'.")

#     # Preparing input and output file(s) / folder(s)
#     if os.path.isfile(project_folder):
#         project_folder = os.path.dirname(project_folder)
#         if debug:
#             print(f"Project folder is set to: {project_folder}")

#     files = sorted([f for f in list_files(project_folder, extension='tsv', recursive=True, ) if not 'tmp' in f])

#     if len(files) == 0:
#         print('No files found. Exiting.')
#     elif len(files) == 1:
#         pool = False

#     if output:
#         project_folder = os.path.join(project_folder, output)
#         make_dir(project_folder)

#     # Running main deduplication process
#     total_sequences = 0
#     pooled = []

#     for file in files:
#         sample = os.path.basename(file).split('.')[0]

#         print(f"Processing {sample}")
#         print("-"*(len(sample)+11)+'\n')

#         # Loading files
#         if umi:
#             try:
#                 df = pl.read_csv(file, columns=['sequence_id', 'sequence', 'umi'], separator='\t', null_values="None", low_memory=True if large_files else False, )
#             except Exception as e:
#                 print(f"Error reading file {file}: {e}")
#                 continue
#         else:
#             try:
#                 df = pl.read_csv(file, columns=['sequence_id', 'sequence'], separator='\t', null_values="None", low_memory=True if large_files else False, )
#             except Exception as e:
#                 print(f"Error reading file {file}: {e}")
#                 continue

#         print(f"Loaded {df.height:,} annotated sequences")
#         total_sequences += df.shape[0]

#         # Processing with deduplication in the presence of UMIs
#         if umi:
#             if debug:
#                 print(f"Deduplicating with UMI ({df.unique(subset=['umi']).height} unique UMIs)")

#             df_unique = _deduplicate_sequences(df, ["sequence", "umi"], keep_read_numbers)


#         # Processing with deduplication in the absence of UMIs
#         else:
#             df_unique = _deduplicate_sequences(df, ["sequence"], keep_read_numbers)


#         msg = f"Found {df_unique.height:,} unique sequences"
#         if pool:
#             msg += " added to the pool\n"
#         print(msg)

#         if pool:
#             pooled.append(df_unique)
#         else:
#             # Saving deduplicated single output to FASTA file
#             if output_format == "fasta":
#                 fasta_file = os.path.join(project_folder, sample)+'.fasta'

#                 _write_fasta(df_unique, fasta_file, keep_read_numbers, read_number_separator)


#             # Saving deduplicated single output to TSV file
#             elif output_format == "airr":
#                 tsv_file = os.path.join(project_folder, sample)+'.tsv'
#                 # Convert df_unique to LazyFrame for more efficient processing
#                 df_unique_lazy = pl.LazyFrame(df_unique)
#                 # Scan the complete original AIRR table
#                 new_df = pl.scan_csv(file, separator='\t', low_memory=True if large_files else False)

#                 # Filter new_df to only include rows with matching sequence_ids
#                 filtered_df = new_df.join(
#                     df_unique_lazy.select(["sequence_id"]),
#                     on="sequence_id",
#                     how="semi"
#                 ).select(new_df.columns)

#                 # Add the 'count' column to the filtered DataFrame (only for matching rows)
#                 if keep_read_numbers:
#                     filtered_df = filtered_df.join(
#                         df_unique_lazy.select(["sequence_id", "count"]),
#                         on="sequence_id",
#                         how="left"
#                     ).with_columns(
#                         pl.col("count").fill_null(0)  # Replace null values with 0 if necessary
#                     ).rename(
#                         {"count": "duplicates"}  # Rename the column
#                     )

#                 # Sink the resulting DataFrame to a TSV file
#                 filtered_df.collect().write_csv(tsv_file, separator='\t')
#                 print(f"Output written to TSV file: {tsv_file}\n")

#     if pool:
#         pool_df = pl.concat(pooled)

#         if not keep_read_numbers:
#             pool_unique = pool_df.unique(subset=["sequence"]).sort("sequence")
#         else:
#             pool_unique = (
#                                 pool_df.group_by("sequence").agg(
#                                     pl.count().alias("count"),
#                                     pl.col("sequence_id").first(),
#                                 )
#                                 .sort("sequence")
#                             )

#         print(f"\nFound {pool_unique.height:,} unique sequences in pooled data")

#         # Saving deduplicated pooled output to FASTA file
#         if output_format == 'fasta':
#             fasta_file = os.path.join(project_folder, "deduplicated_pool.fasta")
#             _write_fasta(pool_unique, fasta_file, keep_read_numbers, read_number_separator)


#         # Saving deduplicated pooled output to TSV file
#         elif output_format == "airr":
#             tsv_file = os.path.join(project_folder, "deduplicated_pool.tsv")

#             # Convert df_unique to LazyFrame for more efficient processing
#             df_unique_lazy = pl.LazyFrame(pool_unique)

#             # Scan and concatenate multiple AIRR tables
#             new_df = pl.concat(
#                 [pl.scan_csv(file, separator="\t", low_memory=True if large_files else False) for file in files]
#             )

#             # Filter new_df to only include rows with matching sequence_ids
#             filtered_df = new_df.join(
#                 df_unique_lazy.select(["sequence_id"]),
#                 on="sequence_id",
#                 how="semi"
#             ).select(new_df.columns)

#             # Add the 'count' column to the filtered DataFrame (only for matching rows)
#             if keep_read_numbers:
#                 filtered_df = filtered_df.join(
#                     df_unique_lazy.select(["sequence_id", "count"]),
#                     on="sequence_id",
#                     how="left"
#                 ).with_columns(
#                     pl.col("count").fill_null(0)  # Replace null values with 0 if necessary
#                 ).rename(
#                     {"count": "duplicates"}  # Rename the column
#                 )

#             # Sink the resulting DataFrame to a TSV file
#             filtered_df.collect().write_csv(tsv_file, separator='\t')
#             print(f"Output written to TSV file: {tsv_file}\n")


#     # Finalizing
#     end = time.time()
#     duration = end - start
#     print(f"Deduplication complete. Total elapsed time: {datetime.timedelta(seconds=int(duration))}.")

#     return


# def reduction(
#     project_folder: str,
#     output: Optional[str] = None,
#     output_format: str = 'fasta',
#     pool: bool = True,
#     umi: bool = False,
#     keep_cluster_sizes: bool = False,
#     cluster_sizes_separator: str = '|',
#     min_cluster_size: int = 3,
#     clustering_threshold: float = 0.975,
#     consentroid: str = 'centroid',
#     large_files: bool = False,
#     debug: bool = False
# ) -> None:
#     """
#     This function takes as an input AIRR-compliant tables (tsv) and proceeds to
#     data reduction by clustering sequences to a high identity threshold.

#     This is specifically designed to handle large files with minimal footprint.
#     Preclustering can be applied to increase performance over large datasets.

#     Parameters:
#     project_folder (str): Path to the project folder containing AIRR-compliant tables.
#     output (str, optional): Subdirectory for output files. Created if non-existent. Defaults to None.
#     output_format (str): Either "fasta" or "airr". Default is "fasta".
#     pool (bool, optional): If True, pool all samples together. Defaults to True.
#     umi (bool, optional): If True, use UMI for clustering. Defaults to False.
#     keep_cluster_sizes (bool, optional): If True, cluster sizes will be added to sequence names. Defaults to False.
#     cluster_sizes_separator (str, optional): Separator for cluster sizes in sequence names. Defaults to '|'.
#     min_cluster_size (int, optional): Minimum cluster size to consider. Defaults to 3.
#     clustering_threshold (float, optional): Identity threshold for clustering. Defaults to 0.975.
#     consentroid (str, optional): Method to determine cluster representative ('centroid' or 'consensus'). Defaults to 'centroid'.
#     large_files (bool, optional): If True, optimize for large files (>100Go). Defaults to False.
#     debug (bool, optional): If True, print debug information. Defaults to False.

#     Returns: None. A fasta file is written to disk.
#     """

#     start = time.time()

#     if umi:
#         cluster_sizes_separator += "umi_count="
#     else:
#         cluster_sizes_separator += "cluster_size="

#     # Assert that output_format is valid
#     if output_format not in ['fasta', 'airr']:
#         raise ValueError("Invalid output_format. Must be 'fasta' or 'airr'.")

#     # Assert that clustering_threshold is valid
#     if not (0 < clustering_threshold <= 1):
#         raise ValueError("Clustering_threshold must be between 0 and 1.")

#     # Assert that consentroid is valid
#     if consentroid not in ['centroid', 'consensus']:
#         raise ValueError("Consentroid must be either 'centroid' or 'consensus'.")

#     # Check for incompatible consentroid and output_format arguments
#     if consentroid == 'consensus' and output_format == 'airr':
#         raise ValueError(
#             "The 'consensus' consentroid method is incompatible with the 'airr' output format. "
#             "Please use 'centroid' or choose 'fasta' as the output format."
#         )

#     # Preparing input and output file(s) / folder(s)
#     if os.path.isfile(project_folder):
#         project_folder = os.path.dirname(project_folder)
#         if debug:
#             print(f"Project folder is set to: {project_folder}")

#     files = sorted([f for f in list_files(project_folder, extension='tsv', recursive=True, ) if not 'tmp' in f])

#     if len(files) == 0:
#         print('No files found. Exiting.')
#     elif len(files) == 1:
#         pool = False

#     if output:
#         project_folder = os.path.join(project_folder, output)
#         make_dir(project_folder)

#     total_sequences = 0

#     if pool:
#         pooled_heavies = []
#         pooled_lights = []

#     # Processing files individually...
#     for file in files:
#         sample = os.path.basename(file).split('.')[0]

#         print(f"Processing {sample}")
#         print("-"*(len(sample)+11)+'\n')

#         keys = ['sequence_id', 'v_gene', 'j_gene', 'locus', 'sequence']
#         if umi:
#             keys.append('umi')

#         df = pl.read_csv(file, columns=keys, separator='\t', null_values="None", low_memory=True if large_files else False, )

#         if not all(col in df.columns for col in keys):
#             print(f"File {file} is missing required columns. Skipping...")
#             continue

#         print(f"Loaded {df.height:,} annotated sequences")
#         total_sequences += df.height

#         heavies = df.filter(pl.col('locus') == 'IGH')
#         lights = df.filter(pl.col('locus') != 'IGH')

#         if pool:
#             pooled_heavies.append(heavies)
#             pooled_lights.append(lights)

#         else:
#             sample_consentroids = []

#             # Process individual heavy chains
#             heavy_consentroids = _process_chain_group(heavies, 'Heavy', umi, min_cluster_size, clustering_threshold, consentroid, cluster_sizes_separator, keep_cluster_sizes, output_format, debug)
#             sample_consentroids.extend(heavy_consentroids)

#             # Process individual light chains
#             light_consentroids = _process_chain_group(lights, 'Lights', umi, min_cluster_size, clustering_threshold, consentroid, cluster_sizes_separator, keep_cluster_sizes, output_format, debug)
#             sample_consentroids.extend(light_consentroids)

#             # Skipping file creation if no concentroids have been generated
#             if not sample_consentroids:
#                 print(f"No clusters found for sample {sample}. Skipping output.\n")
#                 continue

#             if output_format == 'fasta':
#                 # Save consentroids to FASTA file
#                 fasta_file = os.path.join(project_folder, sample)+'_reduced.fasta'
#                 to_fasta(sample_consentroids, fasta_file)

#             elif output_format == 'airr':
#                 # Saving centroids to TSV file (consensus can't be exported as AIRR-annotated TSV table as they haven't been annotated yet)
#                 tsv_file = os.path.join(project_folder, sample + '_reduced.tsv')

#                 # Convert centroids to LazyFrame for more efficient processing
#                 centroids_lazy = pl.LazyFrame(sample_consentroids, )

#                 # Scan the complete original AIRR table
#                 new_df = pl.scan_csv(file, separator='\t', low_memory=True if large_files else False)

#                 # Filter new_df to only include rows with matching sequence_ids
#                 filtered_df = new_df.join(
#                     centroids_lazy.select(["sequence_id"]),
#                     on="sequence_id",
#                     how="semi"
#                 ).select(new_df.columns)

#                 # Add the 'count' column to the filtered DataFrame (only for matching rows)
#                 if keep_cluster_sizes:
#                     filtered_df = filtered_df.join(
#                         centroids_lazy.select(["sequence_id", "count"]),
#                         on="sequence_id",
#                         how="left"
#                     ).with_columns(
#                         pl.col("count").fill_null(0)  # Replace null values with 0 if necessary
#                     ).rename(
#                         {"count": "umi_count" if umi else "cluster_size"}  # Rename the column
#                     )

#                 # Sink the resulting DataFrame to a TSV file
#                 filtered_df.collect().write_csv(tsv_file, separator='\t')
#                 print(f"Output written to TSV file: {tsv_file}\n")

#     if pool:
#         heavies = pl.concat(pooled_heavies)
#         lights = pl.concat(pooled_lights)

#         all_consentroids = []

#         # Processing pooled heavy chains
#         heavy_consentroids = _process_chain_group(heavies, 'Heavy', umi, min_cluster_size, clustering_threshold, consentroid, cluster_sizes_separator, keep_cluster_sizes, output_format, debug)
#         all_consentroids.extend(heavy_consentroids)

#         # Processing pooled light chains
#         light_consentroids = _process_chain_group(lights, 'Lights', umi, min_cluster_size, clustering_threshold, consentroid, cluster_sizes_separator, keep_cluster_sizes, output_format, debug)
#         all_consentroids.extend(light_consentroids)

#         if output_format == 'fasta':
#             # Saving pooled consentroids to FASTA file
#             fasta_file = os.path.join(project_folder, "reduced_pool.fasta")
#             to_fasta(all_consentroids, fasta_file)

#         elif output_format == 'airr':
#             # Saving consentroids to TSV file
#             tsv_file = os.path.join(project_folder, 'reduced_pool.tsv')

#             # Convert centroids to LazyFrame for more efficient processing
#             centroids_lazy = pl.LazyFrame(all_consentroids)

#             # Scan the complete original AIRR table
#             new_df = pl.concat([pl.scan_csv(f, separator='\t',low_memory=True if large_files else False) for f in files])

#             # Filter new_df to only include rows with matching sequence_ids
#             filtered_df = new_df.join(
#                 centroids_lazy.select(["sequence_id"]),
#                 on="sequence_id",
#                 how="semi"
#             ).select(new_df.columns)

#             # Add the 'count' column to the filtered DataFrame (only for matching rows)
#             if keep_cluster_sizes:
#                 filtered_df = filtered_df.join(
#                     centroids_lazy.select(["sequence_id", "count"]),
#                     on="sequence_id",
#                     how="left"
#                 ).with_columns(
#                     pl.col("count").fill_null(0)  # Replace null values with 0 if necessary
#                 ).rename(
#                     {"count": "umi_count" if umi else "cluster_size"}  # Rename the column
#                 )

#             # Sink the resulting DataFrame to a TSV file
#             filtered_df.collect().write_csv(tsv_file, separator='\t')
#             print(f"Output written to TSV file: {tsv_file}\n")


#     end = time.time()
#     duration = end - start
#     print(f"Data reduction complete. Total elapsed time: {datetime.timedelta(seconds=int(duration))}.")

#     return



#!usr/env/python
# filename: test_utilities.py

from ..utils import utilities


def test_nested_dict_lookup():
    d = {"1a": {"2a": "2a"}, "1b": {"2b": {"3b": "3b"}}}
    a = utilities.nested_dict_lookup(d, ["1a", "2a"])
    b = utilities.nested_dict_lookup(d, ["1b", "2b", "3b"])
    assert a == "2a" and b == "3b"



#!/usr/bin/env python
# filename: phylo.py


#
# Copyright (c) 2023 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os
import subprocess as sp
import tempfile
import uuid
from collections import Counter
from copy import deepcopy
from io import StringIO
from typing import Callable, Iterable, List, Optional, Union

import baltic as bt
import matplotlib.pyplot as plt

# from abstar.core.germline import get_imgt_germlines
from Bio import Phylo
from matplotlib import markers
from matplotlib.path import Path

from ..bin import get_path as get_binary_path
from ..core.sequence import Sequence
from ..io import make_dir, read_fasta, to_fasta
from ..tools.alignment import mafft
from ..tools.cluster import Cluster, cluster

# from ..utils.pipeline import make_dir

__all__ = ["fasttree", "Phylogeny", "phylogeny"]


def fasttree(
    aln: str,
    tree_file: Optional[str] = None,
    is_aa: bool = False,
    fasttree_bin: Optional[str] = None,
    debug: bool = False,
    quiet: bool = True,
) -> str:
    """
    Computes a tree file from a multiple seqeunce alignment using `FastTree`_.

    Parameters
    ----------
    aln : str
        Path to a multiple sequence alignment file, in FASTA format, or a
        FASTA-formatted multiple sequence alignment string. Required.

    tree_file : str
        Path to the tree file which will be output by FastTree. If the parent
        directory does not exist, it will be created. If not provided, the output
        (a Newick-formatted tree file) will be returned as a ``str``.

    is_aa : bool, default=False
        Must be set to ``True`` if the input multiple sequence alignment contains
        amino acid sequences. Default is ``False``, meaning FastTree will expect
        nucleotide sequences.

    fasttree_bin : str, optional
        Path to the desired FastTree binary. Default is to use the version of
        FastTree that is bundled with ``abutils``.

    debug : bool, default=False
        If ``True``, verbose output is printed. Default is False.

    quiet : bool, default=True
        Depricated, but retained for backwards compatibility. Use `debug` instead.


    Returns
    -------
    tree_file: str
        Path to the tree file produced by FastTree.


    .. _FastTree:
        http://www.microbesonline.org/fasttree/
    """
    # process input
    if os.path.isfile(aln):
        alignment_file = os.path.abspath(aln)
    else:
        ff = tempfile.NamedTemporaryFile(delete=False)
        ff.close()
        alignment_file = ff.name
        with open(alignment_file, "w") as f:
            f.write(aln)
    if not quiet:
        debug = True
    # set the FastTree binary
    if fasttree_bin is None:
        fasttree_bin = get_binary_path("fasttree")
        # mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        # # fasttree_bin = os.path.join(
        # #     mod_dir, f"bin/fasttree_{platform.system().lower()}"
        # # )
        # system = platform.system().lower()
        # machine = platform.machine().lower()
        # fasttree_bin = os.path.join(mod_dir, f"bin/fasttree_{system}_{machine}")
        # fasttree_bin = fasttree_bin.replace("x86_64", "amd64")
    # make output directory if necessary
    if tree_file is None:
        as_file = False
        tree_file = tempfile.NamedTemporaryFile(delete=False).name
    else:
        as_file = True
        tree_file = os.path.abspath(tree_file)
    if not os.path.isdir(os.path.dirname(tree_file)):
        make_dir(os.path.dirname(tree_file))
    # run FastTree
    if is_aa:
        ft_cmd = f"{fasttree_bin} {alignment_file} > {tree_file}"
    else:
        ft_cmd = f"{fasttree_bin} -nt {alignment_file} > {tree_file}"
    ft = sp.Popen(ft_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    stdout, stderr = ft.communicate()
    if debug:
        print(ft_cmd)
        print(stdout)
        print(stderr)
    # output
    if as_file:
        return tree_file
    else:
        with open(tree_file, "r") as f:
            tree_string = f.read()
        return tree_string


class Phylogeny:
    """
    Base phylogeny class
    """

    def __init__(
        self,
        sequences: Iterable[Sequence],
        name: Optional[str] = None,
        root: Optional[Union[str, Sequence]] = None,
        cluster: bool = True,
        clustering_threshold: float = 1.0,
        clustering_algo: str = "auto",
        rename: Optional[Union[dict, Callable]] = None,
        id_key: Optional[str] = None,
        sequence_key: Optional[str] = None,
    ):
        """
        Phylogenetic representation of an antibody lineage.

        Parameters
        ----------
        sequences : list of Sequence
            A list of ``abutils.Sequence`` objects. Required.

        name : str, default=None
            Name of the lineage. If not provided, a random name will be generated
            using ``uuid.uuid4()``.

        root : str or Sequence, default=None
            Root of the phylogenetic tree. Can be either a sequence ID (the root sequence must
            be in `sequences`) or an ``abutils.Sequence``. If the provided ``Sequence`` is
            already in `sequences`, the duplicate will be ignored when constructing the tree.
            If not provided, the germline V-gene will be used as root.

        cluster : bool, default=True
            Whether or not to cluster seqeunces by identity prior to alignment and tree inference.

        clustering_threshold : float, default=1.0
            Identity threshold for clustering. Must be between 0-1. Default is ``1.0``, which
            collapses identical sequences.

        rename : dict or Callable, default=None
            Used to rename `sequences`. Can be either a ``dict`` of the format:
                ``{old_name: new_name, ...}``
            or a callable function that accepts the old name and returns the new name. Names
            not found in the ``dict`` or for which the function returns ``None`` will not be
            renamed. If not provided, sequences are not renamed.

        id_key : str, default=None
            Key to retrieve the sequence ID. If not provided or missing, ``Sequence.id`` is used.

        sequence_key : str, default=None
            Key to retrieve the sequence. If not provided or missing, ``Sequence.sequence`` is used.

        """
        self.id_key = id_key
        self.seq_key = sequence_key
        self.name = name if name is not None else str(uuid.uuid4())
        self.do_clustering = cluster
        self.clustering_threshold = clustering_threshold
        self.clustering_algo = clustering_algo
        self._root = root
        self._rename = rename
        self._fasta_string = None
        self._aln_string = None
        self._tree_string = None
        self._germ_db = None
        self._tree = None
        self._sizes = None
        self._clusters = None
        self._cluster_dict = None
        # wait until all other attributes are set before processing sequences
        self.sequences = self.process_sequences(sequences)

    def __eq__(self, other):
        """
        ``Phylogeny`` class instances are equal if their names are the same.
        """
        return self.name == other.name

    def __iter__(self):
        """
        Iterates through ``self.tree.Objects``.
        """
        if self._tree is None:
            return None
        for o in self.tree.Objects:
            yield o

    @property
    def tree(self) -> bt.tree:
        """
        Baltic ``Tree`` object
        """
        if self._tree is None:
            tree_file = StringIO(self.tree_string)
            self._tree = bt.loadNewick(tree_file, tip_regex="_([0-9\-]+)$")
        return self._tree

    @property
    def root(self) -> Union[bt.leaf, Sequence, None]:
        """
        Root of the tree.
        """
        if self._root is None:
            top_v = Counter(
                [s.get("v_call", None) for s in self.sequences]
            ).most_common()[0][0]
            if top_v is None:
                return
            # germ = get_imgt_germlines(self.germ_db, "V", gene=top_v)
            self._root = self._get_top_germline_v()
        elif isinstance(self._root, str):
            root = [s for s in self.sequences if s.id == self._root]
            if not root:
                err = f"WARNING: provided root name ({self._root}) was not found in "
                err += "the list of input sequences. Using the germline V-gene as root."
                print(err)
                self._root = self._get_top_germline_v()
            else:
                self._root = root[0]
        elif self._root is False:
            return None
        return self._root

    @root.setter
    def root(self, root):
        if isinstance(root, str):
            r = [s for s in self.sequences if s.id == root]
            if not r:
                err = f"WARNING: provided root name ({r}) was not found in "
                err += "the list of input sequences. Root was not updated."
                print(err)
            else:
                self._root = r[0]
        else:
            self._root = Sequence(root)

    @property
    def fasta_string(self) -> str:
        if self._fasta_string is None:
            if self.do_clustering:
                sequences = self.clusters.centroids
            if self.root is not None:
                sequences = [self.root] + sequences
            self._fasta_string = to_fasta(sequences)
        return self._fasta_string

    @fasta_string.setter
    def fasta_string(self, fasta_string):
        self._fasta_string = fasta_string
        self._clusters = None
        self._aln_string = None
        self._tree_string = None

    @property
    def aln_string(self) -> str:
        """
        Multiple sequence alignment, as a FASTA-formatted string.
        """
        if self._aln_string is None:
            aln = mafft(self.fasta_string, as_string=True)
            self._aln_string = aln
        return self._aln_string

    @aln_string.setter
    def aln_string(self, aln_string):
        self._aln_string = aln_string
        self._tree_string = None

    @property
    def tree_string(self) -> str:
        """
        Newick tree file, as a string.
        """
        if self._tree_string is None:
            tree_string = fasttree(self.aln_string)
            # reroot only if the root sequence exists
            if self.root is not None:
                tree = Phylo.read(StringIO(tree_string), "newick")
                tree.root_with_outgroup(self.root.id)
                tree_string = tree.format("newick")
            self._tree_string = tree_string
        return self._tree_string

    @tree_string.setter
    def tree_string(self, tree_string):
        self._tree_string = tree_string

    @property
    def germ_db(self) -> str:
        """
        Identifies the germline database used to annotate the sequences. If
        multiple databases were used, the most common one is selected.
        """
        if self._germ_db is None:
            db_counts = Counter(
                [s.get("germline_database", None) for s in self.sequences]
            )
            db = db_counts.most_common()[0][0]
            self._germ_db = db
        return self._germ_db

    @germ_db.setter
    def germ_db(self, germ_db):
        self._germ_db = germ_db

    @property
    def sizes(self) -> dict:
        """
        Returns a ``dict`` of tip sizes using clustering results. Only clusters
        with more than one sequence will be in the ``dict``.
        """
        if self._sizes is None:
            self._sizes = {}
        return self._sizes

    @property
    def clusters(self) -> Iterable[Cluster]:
        if self._clusters is None:
            self.cluster()
        return self._clusters

    @clusters.setter
    def clusters(self, clusters):
        self._clusters = clusters
        self._cluster_dict = None
        self._fasta_string = None
        self._aln_string = None
        self._tree_string = None

    @property
    def cluster_dict(self) -> dict:
        """
        Returns a dict mapping cluster centroid IDs to a list of
        all sequence IDs in the corresponding cluster.
        """
        if self._cluster_dict is None:
            d = {c.centroid.id: c.seq_ids for c in self.clusters}
            self._cluster_dict = d
        return self._cluster_dict

    def sanitize_name(self, name: str) -> str:
        """
        Sanitizes sequence names to be used in the tree, by replacing characters
        that are not allowed in Newick tree files with tokens that can be used to
        un-sanitize the names later.

        Characters that are not allowed in Newick tree files are:
            - colon (:)
            - semicolon (;)
            - comma (,)
            - space ( )
            - open parenthesis (()
            - close parenthesis ())
            - open bracket ([)
            - close bracket (])
            - single quote (')
        """
        return (
            name.replace(":", "<colon>")
            .replace(";", "<semi>")
            .replace(",", "<comma>")
            .replace(" ", "<blank>")
            .replace("(", "<open>")
            .replace(")", "<close>")
            .replace("[", "<openbracket>")
            .replace("]", "<closebracket>")
            .replace("'", "<quote>")
        )

    def unsanitize_name(self, name: str) -> str:
        """
        Un-sanitizes sequence names to be used in the tree, by replacing tokens
        with the original characters.
        """
        return (
            name.replace("<colon>", ":")
            .replace("<semi>", ";")
            .replace("<comma>", ",")
            .replace("<blank>", " ")
            .replace("<open>", "(")
            .replace("<close>", ")")
            .replace("<openbracket>", "[")
            .replace("<closebracket>", "]")
            .replace("<quote>", "'")
        )

    def process_sequences(self, sequences: List[Sequence]) -> List[Sequence]:
        """
        Processes sequences to be used in the tree. Sequence names are sanitized.

        Parameters
        ----------
        sequences : list of Sequence objects
            Sequences to be processed.

        Returns
        -------
        sequences : list of Sequence objects
            Processed sequences.
        """
        # sequences = deepcopy(sequences)
        sequences = [
            Sequence(
                s[self.seq_key] if self.seq_key is not None else s.sequence,
                id=self.sanitize_name(
                    self.rename(s[self.id_key] if self.id_key is not None else s.id)
                ),
            )
            for s in sequences
        ]
        return sequences

    def rename(self, old: str) -> str:
        """
        Sequence renamer. The ``old`` name is sanitized before being passed to
        the ``rename`` function. If ``rename`` is a ``dict``, the unsanitized name
        is used as the key to look up the new name. If ``rename`` is a ``callable``,
        the unsanitized name is passed to the function and the return value is used
        as the new name. If ``rename`` is ``None``, the unsanitized name is returned.
        """
        old = self.unsanitize_name(old)
        if self._rename is None:
            return old
        if callable(self._rename):
            new = self._rename(old)
        else:
            new = self._rename.get(old, None)
        return new if new is not None else old

    def get_size(
        self,
        name: str,
        size: Union[Callable, dict, int, float, None] = None,
        min_size: Union[int, float] = 1,
        default: Union[int, float] = 1,
        multiplier: Union[int, float] = 1,
    ) -> Union[int, float]:
        """
        Get the size of a leaf marker, based on the provided ``size`` argument.

        Parameters
        ----------
        name : str
            Name of the leaf node.

        size : callable, dict, int, float, or None
            If ``callable``, it should take a single argument (the name of the
            leaf node) and return a size. If ``dict``, it should map leaf node
            names to sizes. If ``int`` or ``float``, it should be the size of
            all leaf nodes. If ``None``, the size of all leaf nodes will be 1.

        min_size : int or float
            Minimum size of a leaf node. If the size of a leaf node is less than
            ``min_size``, it will be set to 0.

        default : int or float
            Default size of a leaf node, if it is not found in the ``size``
            argument.

        multiplier : int or float
            Multiplier for the size of a leaf node.

        Returns
        -------
        size : int or float
            Size of the leaf node.
        """
        unsanitized_name = self.unsanitize_name(name)
        if callable(size):
            s = size(unsanitized_name)
        elif isinstance(size, dict):
            s = size.get(unsanitized_name, default)
        elif isinstance(size, (int, float)):
            s = size
        else:
            s = self.sizes.get(name, default)
        if s >= min_size:
            return s * multiplier
        return 0

    def get_color(
        self,
        name: str,
        color: Union[Callable, dict, str, Iterable, None] = None,
        default: Union[str, Iterable] = "black",
    ) -> Union[str, Iterable]:
        """
        Get the color of a leaf marker, based on the provided ``color`` argument.
        If the provided name is a cluster centroid, all sequences in the corresponding
        cluster will be evaluated and the most common color will be used.

        Parameters
        ----------
        name : str
            Name of the leaf node.

        color : callable, dict, str, iterable, or None
            If ``callable``, it should take a single argument (the name of the
            leaf node) and return a color. If ``dict``, it should map leaf node
            names to colors. If ``str``, it should be the color of all leaf nodes.
            If ``iterable``, it should be the color of all leaf nodes, as an
            iterable of RGB(A) values. If ``None``, the color of all leaf nodes
            will be `default`.

        default : str or iterable
            Default color of a leaf node, if it is not found in the ``color``.

        Returns
        -------
        color : str or iterable
            The color of the marker, as a string or an iterable of RGB(A) values.
        """
        names = self.cluster_dict.get(name, [name])
        colors = []
        for n in names:
            n = self.unsanitize_name(n)
            # color is a function
            if callable(color):
                c = color(n)
            # color is a dict
            elif isinstance(color, dict):
                c = color.get(n, default)
            # a single color, as a string
            elif isinstance(color, str):
                c = color
            # a single color, as an iterable of RGB(A) values
            elif (
                isinstance(color, (list, tuple))
                and 3 <= len(color) <= 4
                and all([isinstance(c, (int, float)) for c in color])
            ):
                c = color
            else:
                c = default
            colors.append(c)
        color = Counter(colors).most_common()[0][0]
        return color

    # TODO: add support for branch color orders
    # useful for things like longitudinal data, where the timepoint of the
    # branch can be used to color by the earliest timepoint in the branch

    def get_branch_color(
        self,
        k: Union[bt.node, bt.leaf],
        color: Union[Callable, dict, str, Iterable, None] = None,
        default: Union[str, Iterable] = "black",
    ) -> Union[str, Iterable]:
        """
        Get the color of a branch, based on the provided ``color`` argument.

        Parameters
        ----------
        k : bt.node or bt.leaf
            The branch to get the color of. If a node, the most common color of
            all child leaves will be returned.

        color : callable, dict, str, iterable, or None
            If ``callable``, it should take a single argument (the name of the
            leaf node) and return a color. If ``dict``, it should map leaf node
            names to colors. If ``str``, it should be the color of all leaf nodes.
            If ``iterable``, it should be the color of all leaf nodes, as an
            iterable of RGB(A) values. If ``None``, the color of all leaf nodes
            will be `default`.

        default : str or iterable
            Default color of a leaf node, if it is not found in the ``color``.

        Returns
        -------
        color : str or iterable
            The color of the branch, as a string or an iterable of RGB(A) values.
        """
        if k.is_node():
            leaves = list(k.leaves)
        else:
            leaves = [k]
        colors = []
        for leaf in leaves:
            if not isinstance(leaf, str):
                leaf = leaf.name
            names = self.cluster_dict.get(leaf, [leaf])
            for name in names:
                c = self.get_color(name, color=color, default=default)
                colors.append(c)
        color = Counter(colors).most_common()[0][0]
        return color

    def get_marker_edgewidth(
        self,
        name: str,
        edgewidth: Union[Callable, dict, int, float, None] = None,
        default: Union[int, float] = 1,
    ) -> Union[int, float]:
        """
        Get the edge width of a leaf marker, based on the provided ``edgewidth`` argument.

        Parameters
        ----------
        name : str
            Name of the leaf node.

        edgewidth : callable, dict, int, float, or None
            If ``callable``, it should take a single argument (the name of the
            leaf node) and return an edge width. If ``dict``, it should map leaf node
            names to edge widths. If ``int`` or ``float``, it should be the edge width of
            all leaf nodes. If ``None``, the edge width of all leaf nodes will be `default`.

        default : int or float
            Default edge width of a leaf node, if it is not found in the ``edgewidth``.

        Returns
        -------
        edgewidth : int or float
            The edge width of the marker.
        """
        name = self.unsanitize_name(name)
        if callable(edgewidth):
            w = edgewidth(name)
        elif isinstance(edgewidth, dict):
            w = edgewidth.get(name, default)
        elif isinstance(edgewidth, (int, float)):
            w = edgewidth
        else:
            w = default
        return w

    def cluster(self):
        """
        Cluster sequences prior to alignment.
        """
        clusters = cluster(
            self.sequences,
            threshold=self.clustering_threshold,
            algo=self.clustering_algo,
            iddef=1,
            id_key=self.id_key,
            seq_key=self.seq_key,
        )
        sizes = {}
        for c in clusters:
            sizes[c.centroid.id] = c.size
        self._clusters = clusters
        self._sizes = sizes

    def plot(
        self,
        size: Union[Callable, dict, int, float, None] = None,
        color: Union[Callable, dict, Iterable, str, None] = None,
        alpha: float = 0.75,
        min_size: int = 1,
        size_multiplier: Union[int, float] = 10,
        linewidth: Union[int, float] = 2,
        color_branches: bool = False,
        x_attr: Callable = lambda k: k.height,
        y_attr: Callable = lambda k: k.y,
        connection_type: str = "baltic",
        radial: bool = False,
        radial_start: float = 0,
        radial_fraction: float = 1.0,
        inward_space: float = 0.1,
        marker: str = "o",
        marker_edgewidth: Union[Callable, dict, int, float] = 0,
        marker_edgecolor: Union[Callable, dict, str, Iterable, None] = None,
        marker_halign: str = "left",
        marker_valign: str = "center",
        figsize: Iterable[Union[int, float]] = [8, 8],
        show: bool = False,
        figfile: Union[str, Path] = None,
        **kwargs,
    ) -> Optional[plt.Axes]:
        """
        Plot the phylogenetic tree.

        Parameters
        ----------
        size : callable or dict or int or float or None, optional
            Size of the markers at leaf edges. If a callable, it should take
            a ``baltic`` leaf as input and return a size. If a ``dict``, it
            should have leaf names as keys and sizes as values. If an ``int``,
            or ``float``, all markers will have the same size. If ``None``, the
            cluster sizes will be used.

        color : callable or dict or iterable or str or None, optional
            Color of the markers at leaf edges. If a callable, it should take
            a ``baltic`` leaf object as input and return a color. If a ``dict``,
            it should have leaf names as keys and colors as values. If a ``str``,
            or an iterable of RGB(A) values, all markers will have the same color.
            If ``None``, all markers will be black.

        alpha : float, optional
            Alpha value for the markers.

        min_size : int, optional
            Minimum size of the markers. Any leaf edges with a size smaller than
            `min_size` will not have a marker.

        size_multiplier : int or float, optional
            Multiplier for the marker sizes.

        linewidth : int or float, optional
            Width of the tree lines.

        color_branches : bool, optional
            Whether to color the branches of the tree. If ``True``, the branches
            will be colored according to the color of the leaf edges.

        x_attr : callable, optional
            Attribute of the ``baltic`` tree object to use for the x-axis.

        y_attr : callable, optional
            Attribute of the ``baltic`` tree object to use for the y-axis.

        connection_type : str, optional
            Type of connection to use. One of ``'baltic'``, ``'direct'``, or ``'elbow'``.

        radial : bool, optional
            Whether to plot the tree in a radial fashion.

        radial_start : float, optional
            Starting point for the radial plot, as a fraction of the circle.

        radial_fraction : float, optional
            Fraction of the circle to use for the radial plot.

        inward_space : float, optional
            Fraction of the circle to leave empty at the center of the radial plot.

        marker : str, optional
            Marker style. See ``matplotlib.pyplot.scatter`` for more details.

         marker_edgewidth : callable or dict or int or float, optional
            Width of the marker edges. If a callable, it should take a ``baltic`` leaf
            as input and return a width. If a ``dict``, it should have leaf names as
            keys and widths as values. If an ``int``, or ``float``, all markers will
            have the same width.

        marker_edgecolor : callable or dict or str or iterable or None, optional
            Color of the marker edges. If a callable, it should take a ``baltic`` leaf
            object as input and return a color. If a ``dict``, it should have leaf names
            as keys and colors as values. If a ``str``, or an iterable of RGB(A) values,
            all markers will have the same color. If ``None``, all marker edges will
            match the marker color.

        marker_halign : str, optional
            Horizontal alignment of the markers. One of ``'left'``, ``'center'``, or
            ``'right'``.

        marker_valign : str, optional
            Vertical alignment of the markers. One of ``'top'``, ``'center'``, or
            ``'bottom'``.

        figsize : iterable of int or float, optional
            Figure size, by default [8, 8]

        show : bool, optional
            Whether to show the figure.

        figfile : str or Path, optional
            Path to save the figure to.

        """
        plt.figure(figsize=figsize)
        ax = plt.gca()

        # plot parameters
        def size_func(k) -> Union[int, float]:
            return self.get_size(
                k.name,
                size=size,
                default=1,
                min_size=min_size,
                multiplier=size_multiplier,
            )

        def color_func(k) -> Union[str, Iterable]:
            return self.get_color(k.name, color=color, default="black")

        def branch_color_func(k) -> Union[str, Iterable]:
            return self.get_branch_color(k, color=color, default="black")

        marker = align_marker(marker, halign=marker_halign, valign=marker_valign)
        if marker_edgecolor is None:
            marker_edgecolor_func = color_func
        else:

            def marker_edgecolor_func(k) -> Union[str, Iterable]:
                return self.get_color(k.name, color=marker_edgecolor, default="black")

        def marker_edgewidth_func(k) -> Union[int, float]:
            return self.get_marker_edgewidth(
                k.name, edgewidth=marker_edgewidth, default=0
            )

        # make the plot
        if radial:
            # circular phlogenetic tree
            inward_space = inward_space * self.tree.treeHeight
            self.tree.plotCircularTree(
                ax,
                x_attr=x_attr,
                y_attr=y_attr,
                width=linewidth,
                colour=branch_color_func if color_branches else None,
                connection_type=connection_type,
                circStart=radial_start,
                circFracn=radial_fraction,
                inwardSpace=inward_space,
            )
            self.tree.plotCircularPoints(
                ax,
                x_attr=x_attr,
                y_attr=y_attr,
                size=size_func,
                colour=color_func,
                zorder=100,
                marker=marker,
                outline_size=marker_edgewidth_func,
                outline_colour=marker_edgecolor_func,
                circStart=radial_start,
                circFrac=radial_fraction,
                inwardSpace=inward_space,
                alpha=alpha,
                **kwargs,
            )
        else:
            # standard phlogenetic tree
            self.tree.plotTree(
                ax,
                x_attr=x_attr,
                y_attr=y_attr,
                width=linewidth,
                colour=branch_color_func if color_branches else None,
                connection_type=connection_type,
            )
            self.tree.plotPoints(
                ax,
                x_attr=x_attr,
                y_attr=y_attr,
                size=size_func,
                colour=color_func,
                zorder=100,
                marker=marker,
                outline_size=marker_edgewidth_func,
                outline_colour=marker_edgecolor_func,
                alpha=alpha,
                **kwargs,
            )
        # style the plot
        for s in ax.spines.keys():
            ax.spines[s].set_visible(False)
        ax.set_xticks([])
        ax.set_yticks([])
        ax.tick_params(length=0, labelsize=0)
        # save or show
        if show:
            plt.tight_layout()
            plt.show()
        elif figfile is not None:
            plt.tight_layout()
            plt.savefig(figfile)
        else:
            return ax

    def _get_top_germline_v(self) -> Union[Sequence, None]:
        """
        Retrieves the germline V-gene sequence. If sequences were assigned different V-gene
        alleles, the most common V-gene allele is used.

        The appropriate germline database is inferred from sequence annotations.

        If the germline V-gene can't be found, the germline database can't be found/inferred,
        or the sequences don't have AIRR-compatible annotations, returns ``None``.
        """
        from abstar.core.germline import get_imgt_germlines

        v_counts = Counter([s.get("v_call", None) for s in self.sequences])
        top_v = v_counts.most_common()[0][0]
        if top_v is None:
            return None
        try:
            germ = get_imgt_germlines(self.germ_db, "V", gene=top_v)
            return Sequence(germ.ungapped_nt_sequence, id=top_v)
        except Exception:
            return


def phylogeny(
    sequences: Union[str, Iterable[Sequence]],
    name: Optional[str] = None,
    root: Optional[Union[str, Sequence]] = None,
    cluster: bool = True,
    clustering_threshold: float = 1.0,
    clustering_algo: str = "auto",
    rename: Optional[Union[dict, Callable]] = None,
    id_key: Optional[str] = None,
    sequence_key: Optional[str] = None,
) -> Phylogeny:
    """
    Phylogenetic representation of an antibody lineage.

    Parameters
    ----------
    sequences : str or list of Sequence
        A list of ``abutils.Sequence`` objects or the path to a FASTA-formatted file.
        Required.

    name : str, default=None
        Name of the lineage. If not provided, a random name will be generated
        using ``uuid.uuid4()``.

    root : str or Sequence, default=None
        Root of the phylogenetic tree. Can be either a sequence ID (the root sequence must
        be in `sequences`) or an ``abutils.Sequence``. If the provided ``Sequence`` is
        already in `sequences`, the duplicate will be ignored when constructing the tree.
        If not provided, the germline V-gene will be used as root.

    cluster : bool, default=True
        Whether or not to cluster seqeunces by identity prior to alignment and tree inference.

    clustering_threshold : float, default=1.0
        Identity threshold for clustering. Must be between 0-1. Default is ``1.0``, which
        collapses identical sequences.

    rename : dict or Callable, default=None
        Used to rename `sequences`. Can be either a ``dict`` of the format:
            ``{old_name: new_name, ...}``
        or a callable function that accepts the old name and returns the new name. Names
        not found in the ``dict`` or for which the function returns ``None`` will not be
        renamed. If not provided, sequences are not renamed.

    id_key : str, default=None
        Key to retrieve the sequence ID. If not provided or missing, ``Sequence.id`` is used.

    sequence_key : str, default=None
        Key to retrieve the sequence. If not provided or missing, ``Sequence.sequence`` is used.

    Returns
    -------
    phylogeny : Phylogeny
        An ``abutils.Phylogeny`` object.
    """
    if isinstance(sequences, str):
        sequences = read_fasta(sequences)
    return Phylogeny(
        sequences=sequences,
        name=name,
        root=root,
        cluster=cluster,
        clustering_threshold=clustering_threshold,
        clustering_algo=clustering_algo,
        rename=rename,
        id_key=id_key,
        sequence_key=sequence_key,
    )


def align_marker(
    marker: str = "o", halign: str = "center", valign: str = "middle"
) -> Path:
    """
    Create markers with specified alignment.
    https://stackoverflow.com/questions/26686722/align-matplotlib-scatter-marker-left-and-or-right


    Parameters
    ----------
    marker : a valid marker specification.
      See mpl.markers

    halign : string, float {'left', 'center', 'right'}
      Specifies the horizontal alignment of the marker. *float* values
      specify the alignment in units of the markersize/2 (0 is 'center',
      -1 is 'right', 1 is 'left').

    valign : string, float {'top', 'middle', 'bottom'}
      Specifies the vertical alignment of the marker. *float* values
      specify the alignment in units of the markersize/2 (0 is 'middle',
      -1 is 'top', 1 is 'bottom').

    Returns
    -------
    marker_array : numpy.ndarray
      A Nx2 array that specifies the marker path relative to the
      plot target point at (0, 0).

    Notes
    -----
    The mark_array can be passed directly to ax.plot and ax.scatter, e.g.::

        ax.plot(1, 1, marker=align_marker('>', 'left'))

    """
    if isinstance(halign, str):
        halign = {
            "right": -1.0,
            "middle": 0.0,
            "center": 0.0,
            "left": 1.0,
        }[halign]
    if isinstance(valign, str):
        valign = {
            "top": -1.0,
            "middle": 0.0,
            "center": 0.0,
            "bottom": 1.0,
        }[valign]
    m = markers.MarkerStyle(marker)
    m_arr = m.get_path().transformed(m.get_transform()).vertices
    m_arr[:, 0] += halign / 2
    m_arr[:, 1] += valign / 2
    return Path(m_arr, m.get_path().codes)
